import pygame
import util
from const import *

from cellulose import *

from OpenGL.GL import *
from texture import Texture

from selectbox import SelectBox as SDLSelectBox

class SelectBox(SDLSelectBox):
    pass


import texture
from cellulose import *
import input
import util
from OpenGL.GL import *
import os


class Input(input.Input):

    def font_clip_rect(self):
        vx = max(self.font.size(self.value)[0] - self.size[0], 0)
        return pygame.Rect((vx, 0), self.size)
    font_clip_rect = ComputedCellDescriptor(font_clip_rect)

    def font_clip_rect(self):
        vx = max(self.font.size(self.value)[0] - self.size[0], 0)
        return (vx, 0, self.size[0], self.size[1])
    font_clip_rect = ComputedCellDescriptor(font_clip_rect)

    def font_value(self):
        surf = self.font.render(self.value, self.style_option["antialias"],
                self.style_option["color"])
        t = texture.Texture(surf)
        if t.width > self.font_clip_rect[2]:
            t = t.sub(*self.font_clip_rect)
        return t
    font_value = ComputedCellDescriptor(font_value)

    def font_y(self):
        return self.pos[1] - self.style_option["padding-top"]
    font_y = ComputedCellDescriptor(font_y)

    def draw_widget(self):
        if self.value:
            glPushMatrix()
            glTranslatef(self.font_x, util.screen_height-self.font_y-self.height, 0)
            self.font_value.render()
            glPopMatrix()

        if self.focused:
            w,h = self.font.size(self.value[0:self.cur_pos])
            x = min(w, self.size[0])
            color = self.style_option["color"]
            color = [c/255.0 for c in color]
            glColor3f(*color)
            glPushMatrix()
            glTranslatef(self.font_x+x, util.screen_height-self.font_y-self.height, 0)
            glBegin(GL_QUADS)
            glVertex2i(0,0)
            glVertex2i(2, 0)
            glVertex2i(2, self.size[1])
            glVertex2i(0, self.size[1])
            glEnd()
            glPopMatrix()
            glColor3f(1,1,1)


import texture
from cellulose import *
import label
import util
from OpenGL.GL import *
import os
import glFreeType


class Label(label.Label):

    def font(self):
        p = os.path.join(util.app.path, self.style_option["font-family"])
        return glFreeType.Font.get(p, self.style_option["font-size"])
    font = ComputedCellDescriptor(font)

    def width(self):
        # Hack to set textlen! REALLY BAD!
        if not self.font.textlen:
            self.font.glPrint (10, 10, self.value)
        w = self.font.textlen
        return util.widgets.Widget.width.function(self, w)
    width = ComputedCellDescriptor(width)

    def height(self):
        h = self.style_option["font-size"]
        return util.widgets.Widget.height.function(self, h)
    height = ComputedCellDescriptor(height)

    def draw_widget(self):
        # FIXME: clipping?
        color = self.style_option["color"]
        color = [c/255.0 for c in color]
        glColor3f(*color)
        self.font.glPrint (self.font_x, util.screen_height-self.font_y-\
                self.height, self.value)
        glColor3f(1,1,1)


from widget import Widget as SDLWidget
from OpenGL.GL import *
from glstyleset import StyleSet
import util
import pygame
from cellulose import *


class Widget(SDLWidget):

    def draw(self, drawbg=True):
        if self.focused:
            self.repetitive_events()

        x,y = self.pos
        y = util.screen_height-y
        if self.container:
            y -= self.container.offset_y
            if self.container.offset_y > self.y:
                return
            if self.container.offset_y+self.container.height < self.y+self.height:
                return
        glPushMatrix()
        glTranslatef(x, y, 0)
        self.style.draw()
        glPopMatrix()
        self.draw_widget()

    def rect(self):
        return pygame.Rect(self.pos[0], self.pos[1], self.width, self.height)
    rect = ComputedCellDescriptor(rect)


import util
from util import glinit as init
util.gl_assign()

# Import from gllabel2 and glinput2 for dirrect GL freetype font rendering.
# It's just buggy and crashes alot... it's also slower! :(
# Only good thing is that it looks *much* better.

from glwidgets import *
from container import Container
from app import App
from button import Button, Switch
from gllabel import Label
from glstyleset import StyleSet
from box import VBox, HBox
from slider import HSlider, VSlider
from gltextblock import TextBlock
from glimage import Image
from glselectbox import SelectBox
from glinput import Input


import pygame, sys, math
from pygame.locals import *
try:
    from OpenGL.GL import *
except ImportError:
    print "Wanning: could not import OpenGL! Will not work if game uses OpenGL!"


class Texture:
    texture_id = None
    def __init__(self, image, internal_format=GL_RGBA):
        if isinstance(image, str):
            self.filename = image
            image = pygame.image.load(image)
        else:
            self.filename = repr(image)
        w, h = image.get_width(), image.get_height()

        # We need to make sure the texture size is always a power of 2.
        tw = int(2 ** math.ceil(math.log(w, 2)))
        th = int(2 ** math.ceil(math.log(h, 2)))

        if w != tw or h != th:
            w,h = tw,th
            s = pygame.Surface((w,h), SRCALPHA, image)
            s.blit(image, (0,h-image.get_height()))
        else:
            s = image
        self.width = w
        self.height = h


        data = pygame.image.tostring(s, 'RGBA', 1)
        self.texture_id = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.texture_id)
        glPixelStorei(GL_UNPACK_ROW_LENGTH, 0)
        try:
            glTexImage2D(GL_TEXTURE_2D, 0, internal_format, w, h,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, data)
        except:
            print "Warning: Texture too big to fit into video memory!"
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)

    def get_rect(self):
        return pygame.Rect(0,0,self.width,self.height)

    def __repr__(self):
        return '<Texture %rx%r %r>'%(self.width, self.height,
            self.filename)

    def __del__(self):
        if self.texture_id is None: return
        try:
            glDeleteTextures(self.texture_id)
            self.texture_id = None
        except NameError:
            # glDeleteTextures has been gobbled
            pass

    def render(self, mode=GL_MODULATE):
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self.texture_id)
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mode)
        glBegin(GL_QUADS)
        glTexCoord2f(0, 0)
        glVertex2f(0, 0)
        glTexCoord2f(1, 0)
        glVertex2f(self.width, 0)
        glTexCoord2f(1, 1)
        glVertex2f(self.width, self.height)
        glTexCoord2f(0, 1)
        glVertex2f(0, self.height)
        glEnd()
        glDisable(GL_TEXTURE_2D)

    def sub(self, x, y, w, h):
        return SubTexture(self, x, y, w, h)


class SubTexture(Texture):
    def __init__(self, texture, x, y, w, h):
        self.texture = texture
        self.filename = texture.filename
        self.x, self.y = x, y
        self.w, self.h = w, h
        self.width, self.height = w, h
        self.texture_id = texture.texture_id

    def render(self, frame=0):
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self.texture.texture_id)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
        glBegin(GL_QUADS)
        w, h = float(self.texture.width), float(self.texture.height)
        su, sv = self.x/w, self.y/h
        eu, ev = su + self.w/w, sv + self.h/h
        glTexCoord2f(su, sv)
        glVertex2f(0, 0)
        glTexCoord2f(eu, sv)
        glVertex2f(self.w, 0)
        glTexCoord2f(eu, ev)
        glVertex2f(self.w, self.h)
        glTexCoord2f(su, ev)
        glVertex2f(0, self.h)
        glEnd()
        glDisable(GL_TEXTURE_2D)

    def __del__(self):
        pass



class TextureGrid(Texture):
    def __init__(self, filename, width, height):
        Texture.__init__(self, filename)
        self.frames = []
        self.frame_width = width
        self.frame_height = height

        for x in range(0, self.width, width):
            for y in range(0, self.height, height):
                x = float(x); y = float(y)
                uvs = [
                    (x / self.width, y / self.height),
                    ((x+width) / self.width, y / self.height),
                    ((x+width) / self.width, (y+height) / self.height),
                    (x / self.width, (y+height) / self.height),
                ]
                self.frames.append(uvs)

    def row(self, n):
        y = float(n * self.frame_height)
        frames = []
        for x in range(0, self.width, self.frame_width):
            x = float(x)
            uvs = [
                (x / self.width, y / self.height),
                ((x+self.frame_width) / self.width, y / self.height),
                ((x+self.frame_width) / self.width,
                    (y+self.frame_height) / self.height),
                (x / self.width, (y+self.frame_height) / self.height),
            ]
            frames.append(uvs)
        return TextureGridRow(self, frames, self.frame_width,
            self.frame_height)

class TextureGridRow(Texture):
    def __init__(self, texture, frames, frame_width, frame_height):
        self.__texture = texture            # don't collect the parent!
        self.filename = texture.filename
        self.width = texture.width
        self.height = texture.height
        self.texture_id = texture.texture_id
        self.frames = frames
        self.frame_width = frame_width
        self.frame_height = frame_height

    def render(self, frame=0):
        uvs = self.frames[frame]
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_BLEND)
        glBindTexture(GL_TEXTURE_2D, self.__texture.texture_id)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
        glBegin(GL_QUADS)
        glTexCoord2f(*uvs[0])
        glVertex2f(0, 0)
        glTexCoord2f(*uvs[1])
        glVertex2f(self.frame_width, 0)
        glTexCoord2f(*uvs[2])
        glVertex2f(self.frame_width, self.frame_height)
        glTexCoord2f(*uvs[3])
        glVertex2f(0, self.frame_height)
        glEnd()
        glDisable(GL_BLEND)
        glDisable(GL_TEXTURE_2D)

    def __del__(self):
        pass



from __future__ import division
import util
import pygame
from cellulose import *
from OpenGL.GL import *
from styleset import StyleSet as SDLStyleSet
from texture import Texture


class Style:
    def __init__(self, styleset, values):
        self.styleset = styleset
        self.widget = styleset.widget
        self.values = values
        self.setup(values)

    def setup(self, values):
        pass

    def render(self):
        pass


class Opacity(Style):
    def setup(self, values):
        if len(values) == 1:
            values = values[0]
        self.values = int(values)
        # TODO: actually set alpha (or maybe do in draw func?).
        self.styleset.applied_styles["opacity"] = self


class BgColor(Style):
    def setup(self, values):
        self.color = None
        if values[0] != "none":
            if type(values[0]) == str:
                color = values[0][1:-1].split(",")
                color = [int(c) for c in color]
            else:
                color = values[0]
            self.color = [c/255.0 for c in color]

        self.styleset.applied_styles["bgcolor"] = self

    def render(self):
        if not self.color: return
        color = self.color
        o = 1
        if len(color) < 4:
            if self.styleset["opacity"]:
                o = self.styleset["opacity"].values/255
                color.append(o)
            else:
                color.append(1)

        y = 0
        if self.widget.container:
            y += self.widget.container.offset_y

        glColor4f(*color)
        glBegin(GL_QUADS)
        glVertex2f(0, y)
        glVertex2f(self.widget.width, y)
        glVertex2f(self.widget.width, y-self.widget.height)
        glVertex2f(0, y-self.widget.height)
        glEnd()
        glColor4f(1,1,1,o)


class Border(Style):
    def setup(self, values):
        color = values[1][1:-1].split(",")
        self.color = [int(c) for c in color]
        self.width = int(values[0])
        self.styleset.applied_styles["border"] = self

    def render(self):
        pass


class BgImage(Style):
    def setup(self, values):
        if values[0] != "none":
            self.surf = util.app.get_image(values[0])
            self.im = Texture(self.surf)
            self.sub = self.im.sub(0,0,self.surf.get_width(), self.surf.get_height())
            self.style = None
            if len(values) > 1:
                self.style = values[1]
        else:
            self.style = "none"
        self.styleset.applied_styles["bgimage"] = self

        if values[0] != "none":
            # Here we slice up the texture into required sections.
            w = self.w = self.surf.get_width()//2
            h = self.h = self.surf.get_height()//2
            self.tl = self.im.sub(0, 0, w, h)
            self.bl = self.im.sub(0, self.surf.get_height()-h, w, h)
            self.tr = self.im.sub(self.surf.get_width()-w, 0, w, h)
            self.br = self.im.sub(self.surf.get_width()-w, self.surf.get_height()-h, w, h)
            self.l = self.im.sub(0, h, w, 2)
            self.r = self.im.sub(self.surf.get_width()-w, h, w, 2)
            self.t = self.im.sub(w, 0, 4, h)
            self.b = self.im.sub(w, self.surf.get_height()-h, 2, h)
            c = self.surf.get_at((w+1, h+1))
            self.c = [i/255 for i in c]


    def render(self):
        if self.style == "none": return
        if not self.style or self.style == "no-repeat":
            vw = self.surf.get_width()
            vh = self.surf.get_height()
            self.draw_bg(vw, vh)
        elif self.style == "repeat-y":
            vw = self.surf.get_width()
            vh = self.widget.height
            self.draw_bg(vw, vh, im=self.sub)
        elif self.style == "repeat-x":
            vw = self.widget.width
            vh = self.surf.get_height()
            self.draw_bg(vw, vh, im=self.sub)
        elif self.style == "repeat":
            vw = self.widget.width
            vh = self.widget.height
            self.draw_bg(vw, vh, im=self.sub)

        elif self.style == "stretch":
            sy = self.widget.height/self.surf.get_height()
            sx = self.widget.width/self.surf.get_width()
            glPushMatrix()
            glTranslatef(0, -self.im.height, 0)
            glScalef(sx,sy,1)
            self.im.render()
            glPopMatrix()

        elif self.style == "slice":

            # Draw center.
            glColor4f(*self.c)
            glBegin(GL_QUADS)
            glVertex2f(0, -3)
            glVertex2f(self.widget.width, -3)
            glVertex2f(self.widget.width, -self.widget.height+6)
            glVertex2f(0, -self.widget.height+6)
            glEnd()
            glColor4f(1,1,1, 1)

            # Draw the top and bottom.
            self.draw_bg(self.widget.width, self.t.height,
                    y=self.widget.height-self.b.height,
                    im=self.t, repeat=True)
            self.draw_bg(self.widget.width, self.b.height,
                    im=self.b, y=-1, repeat=True)

            # Left and right
            self.draw_bg(self.l.width, self.widget.height, im=self.l, repeat=True)
            self.draw_bg(self.r.width, self.widget.height,
                    x=self.widget.width-self.r.width, im=self.r, repeat=True)

            # Corners.
            self.draw_bg(self.tl.width, self.tl.height, y=self.widget.height-self.h, im=self.tl, repeat=True)
            self.draw_bg(self.tr.width, self.tr.height,
                    x=self.widget.width-self.w, im=self.tr, repeat=True,
                    y=self.widget.height-self.h)
            self.draw_bg(self.bl.width, self.bl.height, y=-1, im=self.bl, repeat=True)
            self.draw_bg(self.br.width, self.br.height, x=self.widget.width-self.w,
                    y=-1, im=self.br, repeat=True)


    def draw_bg(self, vw, vh, x=0, y=0, im=None, repeat=False):
        # vw and vh are the width and height of the area you want to cover.
        if not im: im = self.im
        vh = -vh
        y = -y

        # We find the location on the texture that we need to get from.
        if repeat:
            w, h = float(im.texture.width), float(im.texture.height)
            su, sv = im.x/w, im.y/h
            eu, ev = su + im.w/w, sv + im.h/h
        else:
            su = sv = 0
            eu = vw/im.width
            ev = vh/im.height

        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, im.texture_id)
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        if repeat:
            # If it's a sub texture, we need to set the wrap to repeat.
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)

        # Draw it.
        glBegin(GL_QUADS)
        glTexCoord2f(su, ev)
        glVertex2f(x, y)

        glTexCoord2f(eu, ev)
        glVertex2f(x+vw, y)

        glTexCoord2f(eu, sv)
        glVertex2f(x+vw, y+vh)

        glTexCoord2f(su, sv)
        glVertex2f(x, y+vh)
        glEnd()
        glDisable(GL_TEXTURE_2D)



class Color(Style):
    def setup(self, values):
        if len(values) != 3:
            color = values[0][1:-1].split(",")
            self.values = [int(c) for c in color]
        else: self.values = values

        self.styleset.applied_styles["color"] = self


class Padding(Style):
    def setup(self, values):
        padding = {}
        if len(values) == 1:
            top = right = bottom = left = values[0]
        elif len(values) == 2:
            top = bottom = values[0]
            left = right = values[1]
        elif len(values) == 4:
            top, right, bottom, left = values

        padding["top"] = int(top)
        padding["right"] = int(right)
        padding["bottom"] = int(bottom)
        padding["left"] = int(left)

        for k in ("top", "left", "right", "bottom"):
            key = "padding-"+k
            if not self.styleset.applied_styles[key]:
                self.styleset.applied_styles[key] = padding[k]



class Effect(Style):
    def setup(self, values):
        if len(values) == 1 and values[0] != "none":
            values.append("10")
        self.styleset.applied_styles["effect"] = values



class StyleSet(SDLStyleSet):
    styles = {
        "opacity":Opacity,
        "bgcolor":BgColor,
        "border":Border,
        "bgimage":BgImage,
        "color":Color,
        "padding":Padding,
        "effect":Effect,
    }

    def draw(self):
        """ Draws the widget based on it's styling. """
        if self.parent:
            self.parent.draw()

        if self["opacity"]:
            o = self["opacity"].values/255
            glColor4f(1,1,1,o)

        for s in self.applied_styles.values():
            if hasattr(s, "render"):
                s.render()
        glColor4f(1,1,1,1)

    def surf(self):
        raise ValueError("GL widgets don't use surface!")
    surf = ComputedCellDescriptor(surf)


    def __getitem__(self, v):
        r = None
        if self.applied_styles.has_key(v):
            r = self.applied_styles[v]
        if not r and self.parent:
            r = self.parent[v]
        if hasattr(r, "value"): r = r.value
        return r

    def __setitem__(self, k, v):
        #self.applied_styles[k] = v
        self.apply(k, v)

    def apply(self, option, values):
        option = option.replace("_", "-")
        o = option.replace("-", "_")
        if type(values) == str or type(values) == int:
            if type(values) == str:
                values = values.split(" ")
            else: values = [values]
        if o in self.styles:
            self.styles[o](self, values)
        else:
            self.generic(option, values)

    def generic(self, n, v):
        if type(v) == list:
            if len(v) == 1:
                v = v[0]
                v = str(v)
                # isdigit returns False if v is a negative number. FIX?
                if v.isdigit(): v = int(v)
        self.applied_styles[n] = v

    #def effect(self, values):
        #if len(values) == 1 and values[0] != "none":
            #values.append("10")
        #self.applied_styles["effect"] = values


import texture
from cellulose import *
import label
import util
from OpenGL.GL import *

class Label(label.Label):

    def font_value(self):
        surf = self.font.render(self.value, self.style_option["antialias"],
                self.style_option["color"])
        return texture.Texture(surf)
    font_value = ComputedCellDescriptor(font_value)

    def draw_widget(self):
        # FIXME: put clipping.
        if self.value:
            glPushMatrix()
            glTranslatef(self.font_x, util.screen_height-self.font_y-self.height, 0)
            self.font_value.render()
            glPopMatrix()


"""
Using GooeyPy with GL!

Note that this is very experimental. The GL render is slow and incomplete. If
are good at GL I would greatly appreciate if you would look at glstyleset.py and
see if you can speed it up. If you do send it in to me and I'll put you on the
credits :)
"""

import sys
sys.path.insert(0, "..")

import pygame
from gooeypy import gl as gui
from gooeypy.const import *

from OpenGL.GL import *

clock = pygame.time.Clock()

vw,vh = viewport = (640, 480)
pygame.display.set_mode(viewport, OPENGL | DOUBLEBUF)


# set up 2d mode
glMatrixMode(GL_PROJECTION)
glLoadIdentity()
glViewport(0, 0, vw, vh)
glOrtho(0, vw, 0, vh, -50, 50)
glMatrixMode(GL_MODELVIEW)
glLoadIdentity()
glDisable(GL_LIGHTING)
glDisable(GL_DEPTH_TEST)

glClearColor(0, 0, 0, 1)
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
glEnable(GL_BLEND)


# Here is one thing we do different. We pass it our own surface that we want it
# to use. This way, it won't set the display.
gui.init(*viewport)

app = gui.App(width=640, height=480)

w1 = gui.Button("reset", x=20, y=30)
w2 = gui.Input(x=100, y=30, width=240)

app.add(w1,w2)


def callback():
    w2.value = "GooeyPy is cool!"
w1.connect(CLICK, callback)

mon = gui.Switch(False, labels=("Multiple", "Single"), x=430, y=200, min_width=110)
app.add(mon)
don = gui.Switch(False, labels=("Disabled", "Enabled"), x=430, y=260, min_width=110)
app.add(don)

sb1 = gui.SelectBox(disabled=don.link("value"), multiple=mon.link("value"),
        x=200, y=200, width=200, height=100, scrollable=True)
sb1.add("Value 1", "value1")
sb1.add("Value 2", "value2")
sb1.add("Value 3", "value3")
sb1.add(w2, "inputvalue")

app.add(sb1)


while True:
    clock.tick(40)

    events = pygame.event.get()

    for event in events:
        if event.type == QUIT:
            sys.exit()

    app.run(events)
    glClear(GL_COLOR_BUFFER_BIT)
    app.draw()
    pygame.display.flip()


import texture
from cellulose import *
import input
import util
from OpenGL.GL import *
import os
import glFreeType

class Input(input.Input):

    def font(self):
        p = os.path.join(util.app.path, self.style_option["font-family"])
        return glFreeType.Font.get(p, self.style_option["font-size"])
    font = ComputedCellDescriptor(font)

    def width(self):
        return util.widgets.Widget.width.function(self)
    width = ComputedCellDescriptor(width)

    def height(self):
        h = self.style_option["font-size"] + self.style_option["padding-top"]+\
                self.style_option["padding-bottom"]
        return util.widgets.Widget.height.function(self, h)
    height = ComputedCellDescriptor(height)

    def font_x(self):
        x,_ = self.pos
        return x + self.style_option["padding-left"]
    font_x = ComputedCellDescriptor(font_x)

    def font_y(self):
        _,y = self.pos
        return y - self.style_option["padding-top"]
    font_y = ComputedCellDescriptor(font_y)

    def draw_widget(self):
        color = self.style_option["color"]
        color = [c/255.0 for c in color]
        glColor3f(*color)
        self.font.glPrint (self.font_x, util.screen_height-self.font_y-\
                self.height, self.value)
        glColor3f(1,1,1)

        #if self.focused:
            #w,h = self.font.size(self.value[0:self.cur_pos])
            #r = pygame.Surface((2,self.size[1]))
            #r.fill(self.style_option["color"])
            #x = min(w, self.size[0])
            #util.blit(r, (self.font_x+x, self.font_y))



# import freetype
# We are going to use Python Image Library's font handling
# From PIL 1.1.4:
import ImageFont
from OpenGL.GL import *
from OpenGL.GLU import *
import os


def is_font_available (ft, facename):
    """ Returns true if FreeType can find the requested face name
        Pass the basname of the font e.g. "arial" or "times new roman"
    """
    if (facename in ft.available_fonts ()):
        return True
    return False


def next_p2 (num):
    """ If num isn't a power of 2, will return the next higher power of two """
    rval = 1
    while (rval<num):
        rval <<= 1
    return rval



def make_dlist (ft, ch, list_base, tex_base_list):
    """ Given an integer char code, build a GL texture into texture_array,
        build a GL display list for display list number display_list_base + ch.
        Populate the glTexture for the integer ch and construct a display
        list that renders the texture for ch.
        Note, that display_list_base and texture_base are supposed
        to be preallocated for 128 consecutive display lists and and
        array of textures.
    """

    # //The first thing we do is get FreeType to render our character
    # //into a bitmap.  This actually requires a couple of FreeType commands:
    # //Load the Glyph for our character.
    # //Move the face's glyph into a Glyph object.
    # //Convert the glyph to a bitmap.
    # //This reference will make accessing the bitmap easier
    # - This is the 2 dimensional Numeric array

    # Use our helper function to get the widths of
    # the bitmap data that we will need in order to create
    # our texture.
    glyph = ft.getmask (chr (ch))
    glyph_width, glyph_height = glyph.size
    # We are using PIL's wrapping for FreeType. As a result, we don't have
    # direct access to glyph.advance or other attributes, so we add a 1 pixel pad.
    width = next_p2 (glyph_width + 1)
    height = next_p2 (glyph_height + 1)


    # python GL will accept lists of integers or strings, but not Numeric arrays
    # so, we buildup a string for our glyph's texture from the Numeric bitmap

    # Here we fill in the data for the expanded bitmap.
    # Notice that we are using two channel bitmap (one for
    # luminocity and one for alpha), but we assign
    # both luminocity and alpha to the value that we
    # find in the FreeType bitmap.
    # We use the ?: operator so that value which we use
    # will be 0 if we are in the padding zone, and whatever
    # is the the Freetype bitmap otherwise.
    expanded_data = ""
    for j in xrange (height):
        for i in xrange (width):
            if (i >= glyph_width) or (j >= glyph_height):
                value = chr (0)
                expanded_data += value
                expanded_data += value
            else:
                value = chr (glyph.getpixel ((i, j)))
                expanded_data += value
                expanded_data += value

    # -------------- Build the gl texture ------------

    # Now we just setup some texture paramaters.
    ID = glGenTextures (1)
    tex_base_list [ch] = ID
    glBindTexture (GL_TEXTURE_2D, ID)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)

    border = 0
    # Here we actually create the texture itself, notice
    # that we are using GL_LUMINANCE_ALPHA to indicate that
    # we are using 2 channel data.
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height,
        border, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, expanded_data )

    # With the texture created, we don't need to expanded data anymore
    expanded_data = None



    # --- Build the gl display list that draws the texture for this character ---

    # So now we can create the display list
    glNewList (list_base + ch, GL_COMPILE)

    if (ch == ord (" ")):
        glyph_advance = glyph_width
        glTranslatef(glyph_advance, 0, 0)
        glEndList()
    else:

        glBindTexture (GL_TEXTURE_2D, ID)

        glPushMatrix()

        # // first we need to move over a little so that
        # // the character has the right amount of space
        # // between it and the one before it.
        # glyph_left = glyph.bbox [0]
        # glTranslatef(glyph_left, 0, 0)

        # // Now we move down a little in the case that the
        # // bitmap extends past the bottom of the line
        # // this is only true for characters like 'g' or 'y'.
        # glyph_descent = glyph.decent
        # glTranslatef(0, glyph_descent, 0)

        # //Now we need to account for the fact that many of
        # //our textures are filled with empty padding space.
        # //We figure what portion of the texture is used by
        # //the actual character and store that information in
        # //the x and y variables, then when we draw the
        # //quad, we will only reference the parts of the texture
        # //that we contain the character itself.
        x = float(glyph_width) / float(width)
        y = float(glyph_height) / float(height)

        # //Here we draw the texturemaped quads.
        # //The bitmap that we got from FreeType was not
        # //oriented quite like we would like it to be,
        # //so we need to link the texture to the quad
        # //so that the result will be properly aligned.
        glBegin(GL_QUADS)
        glTexCoord2f(0,0), glVertex2f(0,glyph_height)
        glTexCoord2f(0,y), glVertex2f(0,0)
        glTexCoord2f(x,y), glVertex2f(glyph_width,0)
        glTexCoord2f(x,0), glVertex2f(glyph_width, glyph_height)
        glEnd()
        glPopMatrix()

        # Note, PIL's FreeType interface hides the advance from us.
        # Normal PIL clients are rendering an entire string through FreeType, not
        # a single character at a time like we are doing here.
        # Because the advance value is hidden from we will advance
        # the "pen" based upon the rendered glyph's width. This is imperfect.
        glTranslatef(glyph_width + 0.75, 0, 0)

        # //increment the raster position as if we were a bitmap font.
        # //(only needed if you want to calculate text length)
        glBitmap(0,0,0,0,glyph_width + 0.75,0,None)

        # //Finnish the display list
        glEndList()

    return

# /// A fairly straight forward function that pushes
# /// a projection matrix that will make object world
# /// coordinates identical to window coordinates.
def pushScreenCoordinateMatrix():
    glPushAttrib(GL_TRANSFORM_BIT)
    viewport = glGetIntegerv(GL_VIEWPORT)
    glMatrixMode(GL_PROJECTION)
    glPushMatrix()
    glLoadIdentity()
    gluOrtho2D(viewport[0],viewport[2],viewport[1],viewport[3])
    glPopAttrib()
    return


# Pops the projection matrix without changing the current
# MatrixMode.
def pop_projection_matrix():
    glPushAttrib(GL_TRANSFORM_BIT)
    glMatrixMode(GL_PROJECTION)
    glPopMatrix()
    glPopAttrib()
    return


class Font:
    fonts = {}

    def get(cls, facename, pixel_height):
        if cls.fonts.has_key((facename, pixel_height)):
            return cls.fonts[(facename, pixel_height)]
        return cls(facename, pixel_height)
    get = classmethod(get)

    def __init__ (self, facename, pixel_height):
        # We haven't yet allocated textures or display lists
        #facename = os.path.join(*facename.split("/"))
        self.m_allocated = False
        self.m_font_height = pixel_height
        self.m_facename = facename
        self.fonts[(facename, pixel_height)] = self
        self.textlen = None

        # Try to obtain the FreeType font
        try:
            ft = ImageFont.truetype (facename, pixel_height)
        except:
            raise ValueError, "Unable to locate true type font '%s'" % (facename)

        # Here we ask opengl to allocate resources for
        # all the textures and displays lists which we
        # are about to create.
        self.m_list_base = glGenLists (128)

        # Consturct a list of 128 elements. This
        # list will be assigned the texture IDs we create for each glyph
        self.textures = [None] * 128

        # This is where we actually create each of the fonts display lists.
        for i in xrange (128):
            make_dlist (ft, i, self.m_list_base, self.textures);

        self.m_allocated = True


        # //We don't need the face information now that the display
        # //lists have been created, so we free the assosiated resources.
        # Note: Don't need this, as python will decref and dealloc the ft for us.
        ft = None
        return

    def glPrint (self, x, y, string):
        """
        # ///Much like Nehe's glPrint function, but modified to work
        # ///with freetype fonts.
        """
        # We want a coordinate system where things coresponding to window pixels.
        #pushScreenCoordinateMatrix()

        # //We make the height about 1.5* that of
        h = float (self.m_font_height) / 0.63

        # If There's No Text
        # Do Nothing
        if not string:
            #pop_projection_matrix()
            return

        # //Here is some code to split the text that we have been
        # //given into a set of lines.
        # //This could be made much neater by using
        # //a regular expression library such as the one avliable from
        # //boost.org (I've only done it out by hand to avoid complicating
        # //this tutorial with unnecessary library dependencies).
        # //Note: python string object has convenience method for this :)
        lines = string.split ("\n")

        glPushAttrib(GL_LIST_BIT | GL_CURRENT_BIT  | GL_ENABLE_BIT | GL_TRANSFORM_BIT)
        #glMatrixMode(GL_MODELVIEW)
        #glDisable(GL_LIGHTING)
        glEnable(GL_TEXTURE_2D)
        #glDisable(GL_DEPTH_TEST)
        #glEnable(GL_BLEND)
        #glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glListBase(self.m_list_base)
        modelview_matrix = glGetFloatv(GL_MODELVIEW_MATRIX)

        # //This is where the text display actually happens.
        # //For each line of text we reset the modelview matrix
        # //so that the line's text will start in the correct position.
        # //Notice that we need to reset the matrix, rather than just translating
        # //down by h. This is because when each character is
        # //draw it modifies the current matrix so that the next character
        # //will be drawn immediatly after it.
        for i in xrange (len (lines)):
            line = lines [i]

            glPushMatrix ()
            glLoadIdentity ()
            glTranslatef (x,y-h*i,0)
            glMultMatrixf (modelview_matrix)

            # //  The commented out raster position stuff can be useful if you need to
            # //  know the length of the text that you are creating.
            # //  If you decide to use it make sure to also uncomment the glBitmap command
            # //  in make_dlist().
            glRasterPos2f(0,0)
            glCallLists(str(line))
            rpos = glGetFloatv(GL_CURRENT_RASTER_POSITION)
            self.textlen = abs(x - rpos[0])
            glPopMatrix()

        glPopAttrib()
        #pop_projection_matrix()
        return

    def release (self):
        """ Release the gl resources for this Face.
            (This provides the functionality of KillFont () and font_data::clean ()
        """
        if (self.m_allocated):
            # Free up the glTextures and the display lists for our face
            glDeleteLists ( self.m_list_base, 128);
            for ID in self.textures:
                glDeleteTextures (ID);
            # Extra defensive. Clients that continue to try and use this object
            # will now trigger exceptions.
            self.list_base = None
            self.m_allocated = False
        return

    def __del__ (self):
        """ Python destructor for when no more refs to this Face object """
        self.release ()
        return


import pygame
from pygame.locals import *
import util
from const import *

from cellulose import *

from textblock import TextBlock as SDLTextBlock
from OpenGL.GL import *
from texture import Texture


class TextBlock(SDLTextBlock):
    def texture(self):
        linenum = 0
        line_w = self.style_option["width"]
        _,line_h = self.font.size("e")
        surf = pygame.Surface((line_w,line_h*len(self.lines)))
        #print (line_w,line_h*len(self.lines))
        #surf.set_colorkey((0,0,0,0))
        color = [i//255 for i in self.style_option["color"]]
        for line in self.lines:
            x = 0
            y = line_h*linenum
            surf.blit(self.font.render(line, self.style_option["antialias"],
                    color), (x,y))
            linenum += 1
        return Texture(surf)
    texture = ComputedCellDescriptor(texture)

    def draw_widget(self):
        # FIXME: Find out how to draw font in GL.
        pass
        #glPushMatrix()
        #glTranslatef(self.font_x, util.screen_height-self.font_y-self.height, 0)
        #self.texture.render()
        #glPopMatrix()


import pygame
from pygame.locals import *
import util
from const import *

from texture import Texture
from OpenGL.GL import *
from cellulose import *


class Image(util.widgets.Widget):

    value = util.ReplaceableCellDescriptor()

    def __init__(self, value, **params):
        util.widgets.Widget.__init__(self, **params)
        if type(value) == str:
            surf = util.app.get_image(value)
        else:
            surf = value
        self.value = Texture(surf).sub(0,0, surf.get_width(), surf.get_height())

    def draw_widget(self):
        x,y = self.pos
        glPushMatrix()
        glTranslatef(x, util.screen_height-y-self.value.height, 0)
        self.value.render()
        glPopMatrix()